from dataclasses import dataclass, field

from ..errors import BindingError


@dataclass
class GenericBinding:
    """Represents a Generic type binding state."""

    ty: type = None
    instances: list = field(default_factory=list)

    @property
    def is_bound(self):
        """True if the GenericBinding is unbound"""
        return self.ty is not None

    @property
    def none_bound(self):
        return len(self.instances) == 0

    @property
    def can_bind_generic(self):
        """True if the GenericBinding can bind to a new arg."""
        return self.is_bound or self.none_bound

    def can_new_arg_bind(self, arg):
        """True if a given arg can be bound to the current GenericBinding context."""
        return not self.is_bound or self.ty == type(arg)

    def try_bind_new_arg(self, arg):
        if self.can_new_arg_bind(arg):
            self.ty = type(arg)
            self.instances.append(arg)
        else:
            raise BindingError(
                f"Generic bound to different types: {self.ty}, but arg is {type(arg)}"
            )


class GenericBindings(dict):
    def __init__(self, generics):
        super().__init__({G: GenericBinding() for G in generics})

    def try_bind_new_arg(self, ann, arg):
        self.setdefault(ann, GenericBinding()).try_bind_new_arg(arg)



from .validate import (
    validate,
    validator,
    ValidatorFunction,
    TypeValidator,
)
from . import errors

__all__ = [
    "validate",
    "validator",
    "TypeValidator",
    "ValidatorFunction",
    "errors",
]



#!/usr/bin/env python
import inspect, logging
from functools import partial, wraps
from typing import (
    Callable,
    Optional,
    Union,
)

from functools import wraps

from ..errors import *
from ..binding import BindCheckerConfig
from .checker import ValidatorFunction
from .validator import TypeValidator

log = logging.getLogger(__name__)


def validator(func: Callable = None, *, base: Optional[type] = None, **config):
    """Decorator to create custom validator functions for use in validated annotations.

    Can optionally take a base type to check against if the validator function fails.

    ```python
    @validator(base=int)
    def has_c_or_int(arg):
        return True if "c" in arg else False
    ```
    """
    if func is None or not callable(func):
        return partial(validator, base=base, config=config)
    else:
        return ValidatorFunction(func, base, config)


def validate(
    target: Callable | type = None,
    *,
    config: Optional[Union[BindCheckerConfig, dict]] = None,
):
    """Decorator to strictly validate function arguments and return values against their annotations.

    Can optionally take a config dict to change the behavior of the validator.

    ```python
    @validate
    def func(a: int, b: str) -> str:
        return b * a

    @validate(config={"strict": False})
    def func(a: int, b: str) -> str:
        return b * a
    ```
    """
    log.debug(f"{target=} {config=}")

    def _validate_function(func, config):
        return wraps(func)(TypeValidator(func, config=config))

    def _validate_class(cls, config):
        cls.__init__.validated_base_cls = cls

        # Wrap __init__ for validation
        V = _validate_function(cls.__init__, config)
        cls.__init__ = V

        # Inspect the __init__ method to find the fields
        type_annotations = getattr(cls.__init__, "__annotations__", {})

        # Skip 'self' and iterate over the other arguments
        for arg in type_annotations:
            # Attempt to retrieve a validator function
            vf = getattr(cls, f"_{arg}", None)
            log.debug(f"VF {arg=} {vf=}")

            if isinstance(vf, ValidatorFunction):
                field_type = type_annotations.get(arg, None)
                if field_type:
                    V.bind_checker.register_custom_validator(field_type, vf)

        return cls

    if isinstance(config, dict):
        if not isinstance(config, BindCheckerConfig):
            config = BindCheckerConfig(**config)
    elif config is not None and not isinstance(config, BindCheckerConfig):
        raise TypeError(
            f"{config=} must be a dict or BindCheckerConfig, not {type(config)}"
        )
    else:
        config = BindCheckerConfig()

    def decorator(func_or_cls):
        if inspect.isclass(func_or_cls):
            log.debug(f"Class {func_or_cls=}")
            return _validate_class(func_or_cls, config)
        elif callable(func_or_cls):
            log.debug(f"Function {func_or_cls=}")
            return _validate_function(func_or_cls, config)
        else:
            raise TypeError("Invalid target for validation")

    if target is None:
        return decorator
    else:
        return decorator(target)



from dataclasses import dataclass
from typing import Any


@dataclass
class BindCheckerConfig(dict):
    strict: bool = True
    implied_lambdas: bool = False
    ret_validation: bool = True
    disabled: bool = False

    use_custom_validators: bool = True

    performance: bool = False
    no_list_check: bool = False
    no_tuple_check: bool = False
    no_dict_check: bool = False

    ignore_generics: bool = False

    def __getitem__(self, __key: Any) -> Any:
        if __key not in self:
            return None
        return super().__getitem__(__key)



import unittest, logging

from lilvali.validate import validate
from lilvali.errors import *


log = logging.getLogger(__name__)


class TestLilvali(unittest.TestCase):
    def test_modes(self):
        @validate(config={"performance": True})
        def performance_func[T](a: list[T]) -> T:
            return a[0]

        self.assertEqual(performance_func([1, 2, 3]), 1)
        with self.assertRaises(ValidationError):
            performance_func(5)

        # should work because performance mode:
        self.assertEqual(performance_func([1, 2, 3.0]), 1)

        with self.assertRaises(TypeError):

            @validate(config="bad config")
            def bad_config_func(a: int) -> int:  # pragma: no cover
                return a

        # tuple check is also disabled with performance mode
        @validate(config={"performance": True})
        def tuple_func(a: tuple[int, str]) -> str:
            return a[1]

        self.assertEqual(tuple_func((1, "a")), "a")
        with self.assertRaises(ValidationError):
            tuple_func((1, 2))

        @validate(config={"performance": True})
        def dict_func(a: dict[int, str]) -> str:
            return a[1]

        self.assertEqual(dict_func({1: "a"}), "a")
        with self.assertRaises(ValidationError):
            dict_func({1: 2})



from ..errors import BindingError, InvalidType, ValidationError

from .checker import BindChecker
from .config import BindCheckerConfig
from .struct import GenericBinding, GenericBindings

__all__ = ["BindChecker", "BindCheckerConfig", "GenericBinding", "GenericBindings"]



from lilvali import validate, validate
from lilvali.errors import *


@validate
def add[T: (int, float)](x: int, y: T) -> int | float:
    return x + y


def main():
    print(f"{add(1, 2)=}")
    print(f"{add(1, 2.0)=}")

    # TODO: Bug: should raise ValidationError
    try:
        print(f"{add(1.0, 2)=}")
    except ValidationError as e:
        print(f"{e=}")
    else:
        raise RuntimeError("Expected ValidationError")


if __name__ == "__main__":
    main()



import random
import unittest
from cProfile import Profile
from pstats import SortKey, Stats

from lilvali import validate, validator
from lilvali.errors import *


def prof_tests_main():
    # Discover and run tests
    loader = unittest.TestLoader()
    suite = loader.discover(
        start_dir="path_to_your_tests_directory"
    )  # Adjust the path as needed
    runner = unittest.TextTestRunner()
    runner.run(suite)


def prof_main():
    @validator(base=int)
    def has_c_or_int(arg):
        return True if "c" in arg else False

    @validate
    def f[T: (int, float)](x: has_c_or_int, y: T) -> int | float:
        if isinstance(x, str):
            x = int(x.split("=")[1])
        return x + y if random.random() < 0.5 else x - y

    S = 0
    for i in range(100000):
        S = f(
            S if random.random() < 0.5 else f"c={S}", 1 if random.random() < 0.5 else -1
        )


def main():
    profiler = Profile()

    profiler.runcall(prof_main)
    profiler.create_stats()
    stats = Stats(profiler)

    stats.strip_dirs()
    stats.sort_stats(SortKey.CALLS)
    stats.print_stats()
    stats.dump_stats("lilvali_profiling.prof")


if __name__ == "__main__":
    main()
